/**
 @file
 mirrorz~ - sampling/looping object by raja
 
 @ingroup    MSP
 */

#include "ext.h"
#include "ext_obex.h"
#include "ext_common.h" // contains CLAMP macro
#include "z_dsp.h"
#include "ext_buffer.h"

// Hermitic Cubic Interp ( courtesy of Alex Harker: http://www.alexanderjharker.co.uk/ )
#define HRMCBINTRP(f, z, a, b, c) ((((0.5*(c - z) + 1.5*(a - b))*f + (z - 2.5*a + b + b - 0.5*c))*f + (0.5*(b - z)))*f + a)

typedef struct _mirrorz {
    t_pxobject obj;
    t_buffer_ref *l_buffer_reference;
    double sr;
    long nchan;
    long ntrnl_mm_bytelen;
    t_bool playC;
    double *ntrnl_mmL;
    double *ntrnl_mmR;
    double *play_headL;
    double *play_headR;
    double *read_headL;
    double *read_headR;
    double phprev;
    double speed;
    t_ptr_int bufvec_pos;
} t_mirrorz;


void mirrorz_perform64(t_mirrorz *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampfrms, long flags, void *userparam);
void mirrorz_dsp64(t_mirrorz *x, t_object *dsp64, short *count, double samplerate, long maxvectorsize, long flags);
void mirrorz_set(t_mirrorz *x, t_symbol *s);
void *mirrorz_new(t_symbol *s);
void mirrorz_free(t_mirrorz *x);
t_max_err mirrorz_notify(t_mirrorz *x, t_symbol *s, t_symbol *msg, void *sender, void *data);
void mirrorz_in1(t_mirrorz *x, long n);
void mirrorz_assist(t_mirrorz *x, void *b, long m, long a, char *s);
void mirrorz_dblclick(t_mirrorz *x);


static t_class *mirrorz_class;
static inline double ease_func(double y1, char updwn, double ramp, long fad)
{ return updwn ? y1*(0.5*(1.0-cos((1.0-(((double)fad)/ramp))*PI))) : y1*(0.5*(1.0-cos((((double)fad)/ramp)*PI))); }

static inline void interp_index
(t_ptr_int pos,t_ptr_int *indx0,t_ptr_int *indx1,t_ptr_int *indx2,t_ptr_int *indx3,t_ptr_int dir,t_ptr_int end)
{
    *indx0 = pos - dir;                                                         //calc of indexes 4 interps
    if(*indx0 < 0) *indx0 = end + *indx0; else if(*indx0 > end) *indx0 = *indx0 - end;
    
    *indx1 = pos; *indx2 = pos + dir;
    if(*indx2 < 0) *indx2 = end + *indx2; else if(*indx2 > end) *indx2 = *indx2 - end;
    
    *indx3 = *indx2 + dir;
    if(*indx3 < 0) *indx3 = end + *indx3; else if(*indx3 > end) *indx3 = *indx3 - end;
    return;
}

void ext_main(void *r)
{
    t_class *c = class_new("mirrorz~", (method)mirrorz_new, (method)mirrorz_free, sizeof(t_mirrorz), 0L, A_SYM, A_DEFLONG, 0);
    
    class_addmethod(c, (method)mirrorz_dsp64, "dsp64", A_CANT, 0);
    class_addmethod(c, (method)mirrorz_set, "set", A_SYM, 0);
    class_addmethod(c, (method)mirrorz_in1, "in1", A_LONG, 0);
    class_addmethod(c, (method)mirrorz_assist, "assist", A_CANT, 0);
    class_addmethod(c, (method)mirrorz_dblclick, "dblclick", A_CANT, 0);
    class_addmethod(c, (method)mirrorz_notify, "notify", A_CANT, 0);
    class_dspinit(c);
    class_register(CLASS_BOX, c);
    mirrorz_class = c;
}


void mirrorz_perform64(t_mirrorz *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampfrms, long flags, void *userparam)
{
    t_double    *phase = ins[0];
    t_double    *end = ins[1];
    t_double    *playCtrl = ins[2];
    t_double    *recCtrl = ins[3];
    t_double    *speed = ins[4];
    t_double    *recInL = ins[5];
    t_double    *recInR = ins[6];
    t_double    *outL = outs[0];
    t_double    *outR = outs[1];
    t_double    *outPh = outs[2];
    
    int           n = sampfrms;
    t_bool        playC;
    t_ptr_int     bi, chan, frames, nc, bufvecpos, fade, indx, indx0, indx1, indx2, indx3, dir;
    t_float       *tab;
    double        phas, phprev, f, frac, playc, recc, spd, oL, oR, xL, xR, oP;
    double        *ntrnl_mmL = x->ntrnl_mmL;
    double        *ntrnl_mmR = x->ntrnl_mmR;
    double        *playL = x->play_headL;
    double        *playR = x->play_headR;
    double        *readL = x->read_headL;
    double        *readR = x->read_headR;
    double        *trackL;
    double        *trackR;
    
    t_buffer_obj    *buffer = buffer_ref_getobject(x->l_buffer_reference);
    
    tab = buffer_locksamples(buffer);
    if (!tab) goto zero;
    
    frames = buffer_getframecount(buffer);
    nc = buffer_getchannelcount(buffer);
    chan = MIN(x->nchan, nc);
    bufvecpos = x->bufvec_pos;
    playC = x->playC;
    phprev = x->phprev;
    spd = x->speed;
    if (spd>=0) dir = 1; else dir = -1; fade = 0;
    
    if(playC)
    {
        readL = ntrnl_mmL; readR = ntrnl_mmR;
        for (bi=0;bi<sampfrms;bi++)
        {
            f = bufvecpos + (bi * spd); indx = f; if(dir>0) frac = f - indx; else frac = indx - f;
            if (indx<0) indx += frames; else if (indx>=frames) indx -= frames;
            interp_index(indx,&indx0,&indx1,&indx2,&indx3,dir,frames);
            xL = HRMCBINTRP(frac, tab[indx0*nc], tab[indx1*nc], tab[indx2*nc], tab[indx3*nc]);
            xR = HRMCBINTRP(frac, tab[indx0*nc+1], tab[indx1*nc+1], tab[indx2*nc+1], tab[indx3*nc+1]);
            *readL++ = xL; *readR++ = xR;
        }
        playL = ntrnl_mmL; playR = ntrnl_mmR;    bufvecpos += (sampfrms*spd);
        if (bufvecpos>=frames) bufvecpos -= frames; else if (bufvecpos<0) bufvecpos += frames;
    }
    
    while (n--)
    {
        phas = *phase++; playc = *playCtrl++; recc = *recCtrl++; spd = *speed++;
        f = phas * frames; if (spd>=0) dir = 1; else dir = -1;
        if ((phprev!=phas)&&(n>(sampfrms*0.125)))
        {
            trackL = readL = playL; trackR = readR = playR;
            for (bi=0;bi<=n;bi++)
            {
                f = f + (bi * spd); indx = f; if(dir>0) frac = f - indx; else frac = indx - f;
                if (indx<0) indx += indx; else if (indx>=frames) indx -= frames;
                interp_index(indx,&indx0,&indx1,&indx2,&indx3,dir,frames);
                xL = HRMCBINTRP(frac, tab[indx0*nc], tab[indx1*nc], tab[indx2*nc], tab[indx3*nc]);
                xR = HRMCBINTRP(frac, tab[indx0*nc+1], tab[indx1*nc+1], tab[indx2*nc+1], tab[indx3*nc+1]);
                *readL++ = ease_func(xL, 0, n, bi) + ease_func(*playL++, 1, n, bi);
                *readR++ = ease_func(xR, 0, n, bi) + ease_func(*playR++, 1, n, bi);
            }
            playL = trackL; playR = trackR;     bufvecpos = f + (n*spd);
            if (bufvecpos>=frames) bufvecpos -= frames; else if (bufvecpos<0) bufvecpos += frames;
        }
        if ((playc!=playC)&&(n>(sampfrms*0.125)))
        {
            fade = n;
            if (playc == 0) { playC=0; }
            else
            {
                readL = ntrnl_mmL; readR = ntrnl_mmR;
                for (bi=0;bi<=n;bi++)
                {
                    f = f + (bi * spd); indx = f; if(dir>0) frac = f - indx; else frac = indx - f;
                    if (indx<0) indx += frames; else if (indx>=frames) indx -= frames;
                    interp_index(indx,&indx0,&indx1,&indx2,&indx3,dir,frames);
                    xL = HRMCBINTRP(frac, tab[indx0*nc], tab[indx1*nc], tab[indx2*nc], tab[indx3*nc]);
                    xR = HRMCBINTRP(frac, tab[indx0*nc+1], tab[indx1*nc+1], tab[indx2*nc+1], tab[indx3*nc+1]);
                    *readL++ = ease_func(xL, 1, fade, n); *readR++ = ease_func(xR, 1, fade, n);
                }
                playL = ntrnl_mmL; playR = ntrnl_mmR;     bufvecpos = f + (n*spd);     playC = 1;
                if (bufvecpos>=frames) bufvecpos -= frames; else if (bufvecpos<0) bufvecpos += frames;
            }
        }
        
                                /*__________Output__________*/
        if(playC>0) { oL = *playL++; oR = *playR++;; }
        else
        {
            if(!fade) { oL=0.; oR=0.; }
            else { oL=ease_func(*playL++,0,fade,n); oR=ease_func(*playR++,0,fade,n); if(n==1)fade=0; }
        }
        
        oP = (bufvecpos + ((sampfrms - n)*spd))/(double)frames;
        if (oP>=frames) oP -= frames; else if (oP<0) oP += frames;
        *outL++ = oL; *outR++ = oR; *outPh++ = oP; phprev = phas;
    }
    x->phprev = phprev;
    x->playC = playC;
    x->bufvec_pos = bufvecpos;
    x->speed = spd;
    x->play_headL = playL;
    x->play_headR = playR;
    x->read_headL = readL;
    x->read_headR = readR;
    buffer_unlocksamples(buffer);
    return;
zero:
    while (n--) *outL++ = 0.0;
}

void mirrorz_set(t_mirrorz *x, t_symbol *s)
{
    if (!x->l_buffer_reference) x->l_buffer_reference = buffer_ref_new((t_object *)x, s);
    else buffer_ref_set(x->l_buffer_reference, s);
}

void mirrorz_in1(t_mirrorz *x, long n) { if (n) x->nchan = MAX(n, 2) - 1; else x->nchan = 1; }


void mirrorz_dsp64(t_mirrorz *x, t_object *dsp64, short *count, double samplerate, long maxvectorsize, long flags)
{
    x->sr = samplerate; x->ntrnl_mm_bytelen = (maxvectorsize * 2) * sizeof(double);
    if (x->ntrnl_mmL == NULL) { x->ntrnl_mmL = (double *) sysmem_newptrclear(x->ntrnl_mm_bytelen); }
    else
    { x->ntrnl_mmL = (double *) sysmem_resizeptrclear(x->ntrnl_mmL, x->ntrnl_mm_bytelen); }
    if (x->ntrnl_mmL == NULL) { error("mirrorz~: cannot allocate internal memory"); return; }
    x->play_headL = x->ntrnl_mmL;
    if (x->ntrnl_mmR == NULL) { x->ntrnl_mmR = (double *) sysmem_newptrclear(x->ntrnl_mm_bytelen); }
    else
    { x->ntrnl_mmR = (double *) sysmem_resizeptrclear(x->ntrnl_mmR, x->ntrnl_mm_bytelen); }
    if (x->ntrnl_mmR == NULL) { error("mirrorz~: cannot allocate internal memory"); return; }
    x->play_headR = x->ntrnl_mmR;
    dsp_add64(dsp64, (t_object *)x, (t_perfroutine64)mirrorz_perform64, 0, NULL);
}

void mirrorz_dblclick(t_mirrorz *x) { buffer_view(buffer_ref_getobject(x->l_buffer_reference)); }

void mirrorz_assist(t_mirrorz *x, void *b, long m, long a, char *s)
{
    if (m == ASSIST_OUTLET)
    {
        switch (a)
        {
            case 0:    sprintf(s,"Left Out");    break;
            case 1:    sprintf(s,"Right Out");    break;
            case 2:    sprintf(s,"Phase Out");    break;
        }
    }
    else
    {
        switch (a)
        {
            case 0:    sprintf(s,"(signal) Phase/Index Input (other)...");    break;
            case 1:    sprintf(s,"Recording Input Left");    break;
            case 2:    sprintf(s,"Recording Input Right");    break;
            case 3:    sprintf(s,"Play Control");    break;
            case 4:    sprintf(s,"Recording Control");    break;
            case 5:    sprintf(s,"Speed");    break;
        }
    }
}

void *mirrorz_new(t_symbol *s)
{
    long chan = 2;
    t_mirrorz *x = object_alloc(mirrorz_class);
    dsp_setup((t_pxobject *)x, 6);
    outlet_new((t_object *)x, "signal");
    outlet_new((t_object *)x, "signal");
    outlet_new((t_object *)x, "signal");
    mirrorz_set(x, s);
    mirrorz_in1(x,chan);
    x->playC = x->bufvec_pos = 0;
    x->speed = 1.; x->phprev = 0.;
    x->obj.z_misc |= Z_NO_INPLACE;
    post("v.002");
    return (x);
}


void mirrorz_free(t_mirrorz *x)
{ dsp_free((t_pxobject *)x); object_free(x->l_buffer_reference); }


t_max_err mirrorz_notify(t_mirrorz *x, t_symbol *s, t_symbol *msg, void *sender, void *data)
{ return buffer_ref_notify(x->l_buffer_reference, s, msg, sender, data); }
